# returns a relaxed model from a given model

from docplex.mp.utils import DocplexLinearRelaxationError

from collections import defaultdict
import six


class LinearRelaxer(object):
    """ This class returns a linear relaxation for a MIP model.

    """

    def __init__(self):
        self._unrelaxables = defaultdict(list)

    def iter_unrelaxables(self):
        return six.iteritems(self._unrelaxables)

    def main_cause(self):
        unrelaxables = self._unrelaxables
        max_urx = -1
        justifier = None
        main_cause = None
        for cause, urxs in six.iteritems(unrelaxables):
            if len(urxs) > max_urx:
                max_urx = len(urxs)
                main_cause = cause
                justifier = urxs[0]
        return main_cause, justifier

    @staticmethod
    def make_relaxed_model(mdl, **kwargs):
        lrx = LinearRelaxer()
        return lrx.linear_relaxation(mdl, **kwargs)

    def linear_relaxation(self, mdl, **kwargs):
        """ Returns a continuous relaxation of the model.

        Variable types are set to continuous (note that semi-xxx variables have their LB set to zero)

        Some constructs are not relaxable, for example, piecewise-linear expressions, SOS sets,
        logical constraints...
        When a model contains at least one of these non-relaxable constructs, a message is printed
        and this method returns None.

        By default, model parameters are not copied. If you want to copy them, pass the
        keyword argument `copy_parameters=True`

        :param mdl: the initial model

        :return: a new model with continuous relaxation, if possible, else None.
        """

        relax_name = kwargs.get('relaxed_name', None)
        verbose = kwargs.get('verbose', True)
        copy_parameters = kwargs.get('copy_parameters', False)

        model_name = mdl.name

        def info(msg):
            if verbose:
                print("* relaxation of model {0}: {1}".format(model_name, msg))

        mdl_class = mdl.__class__
        unrelaxables = defaultdict(list)

        def process_unrelaxable(urx_, reason):
            unrelaxables[reason or 'unknown'].append(urx_)

        relax_model_name = relax_name or "lp_%s" % mdl.name
        relaxed_model = mdl_class(name=relax_model_name)

        # transfer kwargs
        relaxed_model._parse_kwargs(mdl._get_kwargs())

        # transfer variable containers
        ctn_map = {}
        for ctn in mdl.iter_var_containers():
            copied_ctn = ctn.copy_relaxed(relaxed_model)
            relaxed_model._add_var_container(copied_ctn)
            ctn_map[ctn] = copied_ctn

        # transfer variables
        var_mapping = {}
        continuous = relaxed_model.continuous_vartype
        for v in mdl.iter_variables():
            cpx_code = v.vartype.get_cplex_typecode()
            if not v.is_generated() or cpx_code == 'C':
                # if v has type semixxx, set lB to 0
                if cpx_code in {'N', 'S'}:
                    rx_lb = 0
                else:
                    rx_lb = v.lb
                copied_var = relaxed_model._var(continuous, rx_lb, v.ub, v.name)
                var_ctn = v._container
                if var_ctn:
                    copied_var._container = ctn_map.get(var_ctn)
                var_mapping[v] = copied_var

        # transfer all non-logical cts
        for ct in mdl.iter_constraints():
            if not ct.is_generated():
                if ct.is_logical():
                    process_unrelaxable(ct, 'logical')
                try:
                    copied_ct = ct.relaxed_copy(relaxed_model, var_mapping)
                    relaxed_model.add(copied_ct)
                except DocplexLinearRelaxationError as xe:
                    process_unrelaxable(xe.object, xe.cause)
                except KeyError as ke:
                    info('failed to relax constraint: {0}'.format(ct))
                    process_unrelaxable(ct, 'key')

        # clone objective
        relaxed_model.objective_sense = mdl.objective_sense
        try:
            relaxed_model.objective_expr = mdl.objective_expr.relaxed_copy(relaxed_model, var_mapping)
        except DocplexLinearRelaxationError as xe:
            process_unrelaxable(urx_=xe.object, reason=xe.cause)
        except KeyError:
            process_unrelaxable(urx_=mdl.objective_expr, reason='objective')

        # clone kpis
        for kpi in mdl.iter_kpis():
            relaxed_model.add_kpi(kpi.copy(relaxed_model, var_mapping))

        if mdl.context:
            relaxed_model.context = mdl.context.copy()

        if copy_parameters:
            # copy parameters is not the default behavior
            # by default, the relaxed copy has a clean, default, parameter set.
            # if verbose:
            #     info("copying initial model parameters to relaxed model")
            nb_copied = 0
            for p1, p2 in zip(mdl.parameters.generate_params(), relaxed_model.parameters.generate_params()):
                if p1.is_nondefault():
                    p2.set(p1.get())
                    nb_copied += 1
            if verbose:
                info("copied {0} initial model parameters to relaxed model".format(nb_copied))

        #
        for sos in mdl.iter_sos():
            unrelaxables['sos'].append(sos)

        self._unrelaxables = unrelaxables
        if unrelaxables:
            nb_unrelaxables = len(unrelaxables)
            main_cause, justifier = self.main_cause()

            print("* model {0}: found {1} un-relaxable elements, main cause is {2} (e.g. {3})"
                  .format(mdl.name, nb_unrelaxables, main_cause, justifier))
            if verbose:
                for cause, urxs in six.iteritems(unrelaxables):
                    print('* reason: {0}: {1} unrelaxables'.format(cause, len(urxs)))
                    for u, urx in enumerate(urxs):
                        if hasattr(urx, "is_generated") and urx.is_generated():
                            s_gen = " [generated]"
                        else:
                            s_gen = ""
                        print('--  {0}: cannot be relaxed: {1!s}{2}'.format(u + 1, urx, s_gen))

            return None
        else:
            # force cplex if any, on docloud nothing to do...
            cpx = relaxed_model.get_cplex(do_raise=False)
            if cpx:
                # force type to LP
                cpx.set_problem_type(0)  # 0 is code for LP.
            # sanity  check...
            assert not relaxed_model._contains_discrete_artefacts()
            assert not relaxed_model._solved_as_mip()
            # ---
            return relaxed_model



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module defines the class :class:`CpoSolverListener` that allows to be warned about different steps of the solve.

Any number of listeners can be added to a solver using the method :meth:`docplex.cp.solver.solver.CpoSolver.add_listener`.
Listeners can also be added on the model object using :meth:`docplex.cp.model.CpoModel.add_listener`

This module also defines some default useful listeners:

 * :class:`AutoStopListener`: Listener that stops the solve if configurable conditions are reached.
 * :class:`DelayListener`: Utility listener that waits some time at each solution found.
 * :class:`SolverProgressPanelListener`: implements a progress panel that appears when the solve is started.
   This panel is based on the package *Tkinter* that is available only in Python 2.7.14 and Python 3.X.

All these listeners are effective if the model provides multiple solutions. They are then more adapted to
optimization problems.

To be able to process multiple solutions, the model should be solved:

 * using the solution iterator given by method :meth:`docplex.cp.model.CpoModel.start_search`,
 * or using the method :meth:`docplex.cp.model.CpoModel.solve`
   but setting the parameter *context.solver.solve_with_start_next* to True.

*New in version 2.8.*

Detailed description
--------------------
"""


###############################################################################
## Listener class
###############################################################################

class CpoSolverListener(object):
    """ Solve listener allows to be warned about different solving steps.

    This class is an 'abstract' class that must be extended by actual listener implementation.
    All method of this class are empty.

    *New in version 2.8.*
    """

    def solver_created(self, solver):
        """ Notify the listener that the solver object has been created.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        pass


    def start_solve(self, solver):
        """ Notify that the solve is started.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        pass


    def end_solve(self, solver):
        """ Notify that the solve is ended.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        pass


    def result_found(self, solver, sres):
        """ Signal that a result has been found.

        This method is called every time a result is provided by the solver. The result, in particular the last one,
        may not contain any solution. This should be checked calling method sres.is_solution().

        This method replaces deprecated method solution_found() that is confusing as result may possibly
        not contain a solution to the model.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
            sres:   Solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`
        """
        pass


    def start_refine_conflict(self, solver):
        """ Notify that the refine conflict is started.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        pass


    def end_refine_conflict(self, solver):
        """ Notify that the refine conflict is ended.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        pass


    def conflict_found(self, solver, cflct):
        """ Signal that a conflict has been found.

        This method is called when a conflict result is found by the solver when method refine_conflict() is called.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
            cflct:  Conflict descriptor, object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`
        """
        pass


    def new_log_data(self, solver, data):
        """ Signal a new piece of log data.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
            data:   New log data as a string
        """
        pass


###############################################################################
## Solver listenet that just log events.
###############################################################################

class LogSolverListener(CpoSolverListener):
    """ Solve listener that just log listener events.

    *New in version 2.8.*
    """

    def __init__(self, prefix=""):
        super(LogSolverListener, self).__init__()
        self.prefix = prefix


    def solver_created(self, solver):
        """ Notify the listener that the solver object has been created.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        print(str(self.prefix) + "Solver created")


    def start_solve(self, solver):
        """ Notify that the solve is started.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        print(str(self.prefix) + "Solve started")


    def end_solve(self, solver):
        """ Notify that the solve is ended.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
        """
        print(str(self.prefix) + "Solve ended")


    def result_found(self, solver, sres):
        """ Signal that a result has been found.

        This method is called every time a result is provided by the solver. The result, in particular the last one,
        may not contain any solution. This should be checked calling method sres.is_solution().

        This method replaces deprecated method solution_found() that is confusing as result may possibly
        not contain a solution to the model.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
            sres:   Solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`
        """
        print(str(self.prefix) + "Result found")


###############################################################################
## Implementation of a solve listener that stops search when no new solution
## is given during a given amount of time
###############################################################################

from threading import Condition

class AutoStopListener(CpoSolverListener):
    """ Solver listener that aborts a search when a predefined criteria is reached.

    *New in version 2.8.*
    """
    def __init__(self, qsc_time=None, qsc_sols=None, min_sols=0, max_sols=None):
        """ Create a new solver listener that aborts the solve if defined criteria are reached.

        Args:
            min_sols: Minimun number of solutions to be found before stopping the solve
            max_sols: Maximum number of solutions after which solve is stopped
            qsc_time: Quiesce time limit. Max time, in seconds, after which solver is stopped if no new solution is found.
            qsc_sols: Quiesce time limit expressed as a number of solutions. Quiesce time limit is computed as this
                      value multiplied by the average time between solutions.
        """
        # Initialize next abort time
        self.min_sols = min_sols
        self.max_sols = max_sols
        self.qsc_time = qsc_time
        self.qsc_sols = qsc_sols
        self.nb_sols = 0                                     # Current number of solutions found
        self.start_time = time.time()
        self.last_sol_time = None                            # Time of the last solution
        # Set time checking active indicator
        self.time_active = self.qsc_time is not None or self.qsc_sols is not None

    def start_solve(self, solver):
        # Store solver
        self.solver = solver
        # Check if a time limit is defined
        if self.time_active:
            self.abort_time = self._compute_next_abort_time()
            # Start time waiting thread
            self.time_active = True
            self.condition = Condition()
            thread = Thread(target=self._waiting_loop)
            thread.start()

    def end_solve(self, solver):
        # Stop time control thread if any
        self._stop_waiting_loop()

    def result_found(self, solver, msol):
        self.nb_sols += 1
        self.last_sol_time = time.time()
        # Check if minimal number of solutions not reached
        if self.nb_sols < self.min_sols:
            return
        # Check if max number of solutions has been reached
        if self.max_sols and self.nb_sols >= self.max_sols:
            # Abort search
            self._stop_waiting_loop()
            self.solver.abort_search()
            return
        # Update next abort time
        if self.time_active:
            self.abort_time = self._compute_next_abort_time()
            with self.condition:
                self.condition.notify()

    def _compute_next_abort_time(self):
        if self.nb_sols < self.min_sols:
            return None
        if self.qsc_sols is not None:
            if self.nb_sols > 0:
                delay = self.qsc_sols * ((time.time() - self.start_time) / self.nb_sols)
                if self.qsc_time is not None:
                    delay = min(delay, self.qsc_time)
                return time.time() + delay
            else:
                if self.qsc_time is not None:
                    return time.time() + self.qsc_time
                return None
        if self.qsc_time is not None:
            return time.time() + self.qsc_time
        return None

    def _waiting_loop(self):
        with self.condition:
            while self.time_active:
                ctime = time.time()
                atime = self.abort_time
                if atime is None:
                    self.condition.wait()
                else:
                    if atime <= ctime:
                        self.time_active = False
                        self.solver.abort_search()
                    else:
                        self.condition.wait(atime - ctime)

    def _stop_waiting_loop(self):
        # Stop time control thread if any
        if self.time_active:
            with self.condition:
                self.time_active = False
                self.condition.notify()


###############################################################################
## Implementation of a solve listener that slow down solutions flow by waiting
## a given delay.
###############################################################################


class DelayListener(CpoSolverListener):
    """ Solver listener that waits a given delay after each solution.

    *New in version 2.8.*
    """
    def __init__(self, delay):
        """ Create a new solver listener that waits a given delay after each solution.

        Args:
            delay: Wait delay in seconds
        """
        self.delay = delay

    def result_found(self, solver, msol):
        time.sleep(self.delay)


###############################################################################
## Implementation of a solve progress panel based on Tkinter
###############################################################################

from threading import Thread
import time
from docplex.cp.utils import CpoNotSupportedException

_UNIT_MULTIPLIER = {'k': 1000, 'K': 1000, 'M': 1000000, 'm': 1000000, 'G': 1000000000, 'g': 1000000000,
                    'b': 1, 'B': 1}

TK_INTER_AVAILABLE = True
try:
    # Python 2 version
    import Tkinter as TK
    import tkMessageBox as messagebox
except:
    try:
        # Python 3 version
        import tkinter as TK
        from tkinter import messagebox as messagebox
    except:
        TK_INTER_AVAILABLE = False

if not TK_INTER_AVAILABLE:

    class SolverProgressPanelListener(CpoSolverListener):
        def __init__(self, parse_log=False):
            raise CpoNotSupportedException("Tkinter is not available in this Python environment")

else:

    class _SolverInfoPanel:
        """ Solver info panel that display objective value(s) and KPIs.

        *New in version 2.8.*
        """
        def __init__(self, master, model, stopfun):
            """
            Create a new solver info panel.

            Args:
                master:  Master TK component
                model:   Model to be solved
                stopfun: Function to be called if stop is requested by user
            """
            # Store attributes
            self.master = master
            self.model = model
            self.stopfun = stopfun
            self.start_time = time.time()

            # Set window title
            master.title("CPO solver infos")

            # Add name of the model
            self.model_label = TK.Label(master, text="Solving '" + self.model.get_name() + "'", font="Helvetica 10 bold")
            self.model_label.grid(row=0, column=0, columnspan=2)
            row = 1

            # Add solve time
            TK.Label(master, text="Run time:").grid(row=row, column=0, sticky=TK.E)
            self.time_text = TK.StringVar()
            self.time_text.set("00:00:00")
            TK.Label(master, textvariable=self.time_text, bg='white', width=10, anchor=TK.E)\
                .grid(row=row, column=1, sticky=TK.E)
            row += 1
            master.after(1000, self._update_time)

            # Add objective value if any
            if model.get_optimization_expression() is None:
                self.objective_value_text = None
            else:
                TK.Label(master, text="Objective value:").grid(row=row, column=0, sticky=TK.E)
                self.objective_value_text = TK.StringVar()
                self.objective_value_text.set("unknown")
                TK.Label(master, textvariable=self.objective_value_text, bg='white', width=10, anchor=TK.E)\
                    .grid(row=row, column=1, sticky=TK.E)
                row += 1

            # Add KPIs if any
            kpis = model.get_kpis()
            self.kpi_value_texts = []
            for k in sorted(kpis.keys()):
                TK.Label(master, text=k + ":").grid(row=row, column=0, sticky=TK.E)
                txt = TK.StringVar()
                txt.set("unknown")
                self.kpi_value_texts.append(txt)
                TK.Label(master, textvariable=txt, bg='white', width=10, anchor=TK.E)\
                    .grid(row=row, column=1, sticky=TK.E)
                row += 1

            # Add stop button
            self.stop_button = TK.Button(master, text="Stop solve", command=self.stopfun)
            self.stop_button.grid(row=row, column=0, columnspan=2)

            # Initialize processing of possible additional infos
            self.infos_texts = {}  # Dictionary of info texts per name
            self.last_row = row

        def notify_solution(self, msol):
            """ Notify progress panel about a new solution
            Args:
                msol: New model solution
            """
            if msol:
                # Update objective value
                if self.objective_value_text:
                    self.objective_value_text.set(', '.join(str(v) for v in msol.get_objective_values()))
                # Update KPIs
                kpis = msol.get_kpis()
                for i, k in enumerate(sorted(kpis.keys())):
                    self.kpi_value_texts[i].set(str(kpis[k]))


        def notify_infos(self, infos):
            """ Notify progress panel about last infos values
            Args:
                infos: Information dictionary
            """
            #print("Current infos: {}".format(infos))
            for k, v in infos.items():
                if k in self.infos_texts:
                    txt = self.infos_texts[k]
                else:
                    # Create a new label for info
                    TK.Label(self.master, text=k + ":").grid(row=self.last_row, column=0, sticky=TK.E)
                    txt = TK.StringVar()
                    self.infos_texts[k] = txt
                    TK.Label(self.master, textvariable=txt, bg='white', width=10, anchor=TK.E)\
                        .grid(row=self.last_row, column=1, sticky=TK.E)
                    self.last_row += 1
                    self.stop_button.grid(row=self.last_row, column=0, columnspan=2)
                txt.set(str(v))


        def _update_time(self):
            """ Update value of time field """
            dur = int(time.time() - self.start_time)
            hours = dur // 3600
            mins  = (dur //60) % 60
            secs  =  dur % 60
            self.time_text.set("{:02d}:{:02d}:{:02d}".format(hours, mins, secs))
            self.master.after(1000, self._update_time)


    class _CpoSolverProgressPanel(Thread):
        """ Main class of the solve progress panel.

        It is implemented in a separate thread to be able to run in parallel from the model solving.
        """
        def __init__(self, solver):
            """ Create a new solver progress panel
            Args:
                solver:  Solver to track
                stopfun: Function to be called if stop is requested by user
            """
            super(_CpoSolverProgressPanel, self).__init__()
            self.solver = solver
            self.model = solver.get_model()
            self.active = True
            self.display_panel = None
            self.initdone = False
            self.last_solution = None
            self.last_infos = {}
            self.info_updates = {}

        def run(self):
            """ Thread main loop
            """
            # Create GUI components
            self.root = TK.Tk()
            self.display_panel = _SolverInfoPanel(self.root, self.model, self._stop_requested_by_user)
            self.root.after_idle(self._poll_updates)
            self.root.protocol("WM_DELETE_WINDOW", self.on_closing)
            self.initdone = True

            # Start mainloop
            self.root.mainloop()

        def notify_solution(self, msol):
            """ Notify progress panel about a new solution
            Args:
                msol: New model solution
            """
            self.last_solution = msol

        def notify_infos(self, infos):
            """ Notify miscellaneous information given in a dictionary
            Args:
                infos: Information dictionary
            """
            for k, v in infos.items():
                ov = self.last_infos.get(k)
                if ov != v:
                    self.last_infos[k] = v
                    self.info_updates[k] = v

        def _stop_requested_by_user(self):
            """ Notify a stop requested by end user through GUI """
            self.active = False
            self.solver.abort_search()

        def start(self):
            """ Start the thread and wait for init completed """
            super(_CpoSolverProgressPanel, self).start()
            while not self.initdone:
                time.sleep(0.01)

        def stop(self):
            """ Request progress panel to stop and wait for thread termination """
            self.active = False
            self.join()

        def on_closing(self):
            """ Procedure called when closing the window """
            if messagebox.askokcancel("Quit", "Do you really want to terminate solving of model '{}' ?".format(self.model.get_name())):
                self._stop_requested_by_user()

        def _poll_updates(self):
            """ Polling of external changes by the TK mainloop """
            # Check stop requested
            if not self.active:
                self.root.quit()
                return
            # Check new solution
            sol = self.last_solution
            if sol is not None:
                self.last_solution = None
                self.display_panel.notify_solution(sol)
            ifoupd = self.info_updates
            if ifoupd:
                self.info_updates = {}
                self.display_panel.notify_infos(ifoupd)
            # Rearm callback
            self.root.after(1000, self._poll_updates)


    class SolverProgressPanelListener(CpoSolverListener):
        """ Solver listener that display a solving progress panel.

        The solver progress panel displays the following information:

         * the time elapsed since the beginning of the solve,
         * the last known objective (if any)
         * the last known values of the KPIs (if any)
         * a *Stop solve* button allowing to stop the solve and keep the last known solution as model solution.
         * if parse_log indicator is set, information taken from the log: memory usage, bounds, etc

        *New in version 2.8.*

        Args:
            parse_log(optional): Enable log parsing to retrieve additional information such as memory, bounds, etc
        """
        def __init__(self, parse_log=False):
            super(SolverProgressPanelListener, self).__init__()
            self.parse_log = parse_log

        def solver_created(self, solver):
            # Force solve with start/next
            solver.context.solver.solve_with_start_next = True


        def start_solve(self, solver):
            """ Notify that the solve is started. """
            # Create progress panel
            self.progress_panel = _CpoSolverProgressPanel(solver)
            self.progress_panel.start()

        def end_solve(self, solver):
            """ Notify that the solve is ended. """
            self.progress_panel.stop()

        def result_found(self, solver, msol):
            """ Signal that a solution has been found. """
            self.progress_panel.notify_solution(msol)

        def new_log_data(self, solver, data):
            """ Signal a new piece of log data. """
            if not self.parse_log:
                return
            # Search for solve infos
            for line in data.splitlines():
                if line.startswith(" ! Time = "):
                    infos = {}
                    ex = line.find("s,", 10)
                    if ex > 0:
                        infos['SolveTime'] = float(line[10:ex])
                    sx = line.find("Average fail depth = ", ex)
                    if sx >= 0:
                        sx += 21
                        ex = line.find(", ", sx)
                        if ex > 0:
                            infos['Average fail depth'] = int(line[sx:ex])
                    sx = line.find("Memory usage = ", ex)
                    if sx >= 0:
                        sx += 15
                        ex = line.find(" ", sx)
                        if ex > 0:
                           #infos['Memory usage'] = int(float(line[sx:ex]) * _UNIT_MULTIPLIER[line[ex+1]])
                           infos['Memory usage'] = line[sx:]
                    self.progress_panel.notify_infos(infos)
                elif line.startswith(" ! Current bound is "):
                    infos = {}
                    ex = line.find(" (", 20)
                    if ex > 0:
                        #infos['Current bound'] = [int(x) for x in line[20:ex].split(";")]
                        infos['Current bound'] = line[20:ex]
                    self.progress_panel.notify_infos(infos)
                elif line.startswith(" ! Current objective is "):
                    infos = {'Current objective': line[24:]}
                    #infos['Current objective'] = [int(x) for x in line[24:].split(";")]
                    self.progress_panel.notify_infos(infos)




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2019-2020
# --------------------------------------------------------------------------

# gendoc: ignore

import json
import uuid
try:
    from IPython.display import Javascript, HTML, display
except ImportError:
    Javascript = None

"""
There are two versions of the segments informations.
Version 1 in the version used until May 2020
Version 2 is the version used after May 2020.

During the transition time, we will generate segments tracking for both versions.
After every infrastructure is in version 2, we will remove version 1.

"""
    
# Those were the initial values
CATEGORY = "Decision Optimization"
START_SOLVE_EVENT = "Solve in Notebook"
END_SOLVE_EVENT = "End solve in Notebook"
# Process values for version-2
PROCESSES = {START_SOLVE_EVENT: 'Started Process',
             END_SOLVE_EVENT: 'Ended Process'}

def generate_payload(category, event, details=None, hw_spec=None, version=1):
    if version==1:
        payload = {'action' : event,
                   'model': {'type': 'python'},
                   'category': category }
        if details:
            payload['details'] = details
        if hw_spec:
            payload['hardware_spec'] = hw_spec
    elif version==2:
        payload = {
            'platformTitle': "Watson Studio",
            'productTitle': "Decision Optimization",
            'processType': "do_solve",
            'process': "do_solve/{modelType}/{solveId}".format(**details),
            'successFlag': True, # success in starting the process
            'custom.modelType': "python",
            'custom.solveId': details['solveId']
        }
        if event == START_SOLVE_EVENT:
            payload.update({'custom.modelSize.numConstraints': details['modelSize']['numConstraints'],
                            'custom.modelSize.numVariables': details['modelSize']['numVariables'],
                            'custom.engineEdition': details['edition'],
                            'custom.engineType': details['modelType'],})
        elif event == END_SOLVE_EVENT:
            payload.update({'custom.solveTime': details['solveTime']})
    else:
        raise ValueError("'version' is not in valid range")
    return json.dumps(payload, indent=3)

def generate_js(category, event, details=None, version=1):
    '''
    For versions:
    
       - 1: refers to the old action type
       - 2: refers to the new action type see https://github.ibm.com/IBMDecisionOptimization/dd-planning/issues/2266
    
    Args:
        version: 1 if older segment type, 2 if new segment type.
    '''
    if version == 1:
        template ='''
            if (parent && parent.analytics)
                parent.analytics.track("{category}: {event}",
                                        {payload})
        '''.format(payload=generate_payload(category, event, details),
                   category=category, event=event)
    elif version == 2:
        template ='''
            if (parent && parent.analytics)
                parent.analytics.track("{process}",
                                        {payload})
        '''.format(payload=generate_payload(category, event, details, version=2),
                   process=PROCESSES[event])

    else:
        raise ValueError("'version' is not in valid range")
    return template


class Tracker(object):    
    # Amplitude tracker to get instrumentation when running on WS
    def notify_ws(self, event, details=None):
        if Javascript:
            js = generate_js(CATEGORY, event, details)
            uid = str(uuid.uuid4())
            display(Javascript(js), display_id=uid)
            display(HTML('<div></div>'), update=True, display_id=uid)
            # send also new segment types
            js = generate_js(CATEGORY, event, details, version=2)
            uid = str(uuid.uuid4())
            display(Javascript(js), display_id=uid)
            display(HTML('<div></div>'), update=True, display_id=uid)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

# This file contains all compatibility stuff between py2/py3
import sys

# py2/py3 compatibility
try:
    from Queue import Queue
except ImportError:
    # noinspection PyUnresolvedReferences
    from queue import Queue

# copy_reg is copyreg in Py3
try:
    import copy_reg as copyreg
except ImportError:  # pragma: no cover
    import copyreg   # pragma: no cover

# we want StringIO to process strings in Py2 and Py3
try:
    from cStringIO import StringIO
except ImportError:  # pragma: no cover
    from io import StringIO  # pragma: no cover

try:
    from string import maketrans as mktrans  # Python 2
except ImportError:                 # pragma: no cover
    def mktrans(a, b):              # pragma: no cover
        return str.maketrans(a, b)  # pragma: no cover


try:
    # noinspection PyCompatibility
    xrange(2)
    fast_range = xrange
except NameError:       # pragma: no cover
    fast_range = range  # pragma: no cover

try:
    from itertools import izip
except ImportError:  # pragma: no cover
    izip = zip       # pragma: no cover

# we want unicode in py2, str otherwise
# unitext() returns a unicode representation of its parameter
if sys.version_info[0] == 3:  # pragma: no cover
    unitext = str
else:
    unitext = unicode

try:  # pragma: no cover
    from itertools import zip_longest as izip_longest
except ImportError:  # pragma: no cover
    from itertools import izip_longest


# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
from docplex.mp.constants import ComparisonType
from docplex.mp.linear import LinearExpr, Var
from docplex.mp.mprinter import TextModelPrinter, _ExportWrapper
from docplex.mp.vartype import ContinuousVarType, IntegerVarType, BinaryVarType


class ModelPrettyPrinter(TextModelPrinter):
    vartype_map = {ContinuousVarType: "float", IntegerVarType: "int", BinaryVarType: "bool"}

    ct_symbol_map = {ComparisonType.EQ: "==",
                     ComparisonType.LE: "<=",
                     ComparisonType.GE: ">="}

    indented = ' ' * 2

    def __init__(self, nb_digits=6, sort_variable_names=False):
        # comment line is // as in OPL
        # do NOT forget about user names
        # no encoding is printed
        TextModelPrinter.__init__(self, indent=2, comment_start='//',
                                  nb_digits_for_floats=nb_digits,
                                  hide_user_names=False,
                                  encoding=None,
                                  sort_variable_names=sort_variable_names)

    def get_format(self):
        from docplex.mp.format import OPL_format

        return OPL_format

    def fix_name(self, mobj, prefix, local_index_map, hide_names):
        raw_name = mobj.name
        if raw_name and not hide_names and not mobj.is_generated():
            # there is a user name
            return self._translate_chars(raw_name)
        elif not isinstance(mobj, Var):
            # constraitn with no name -> no name
            return None
        elif mobj.is_generated():
            #mobj_origin = mobj.origin()
            # if hasattr(mobj_origin, 'as_var') and mobj is mobj_origin.as_var:
            #     return str(mobj_origin)
            # else:
            return self._make_prefix_name(mobj, prefix, local_index_map, offset=1)
        else:
            # anonymous or anonymized variable. -> x<i>
            return self._make_prefix_name(mobj, prefix, local_index_map, offset=1)

    def _print_model_name(self, out, mdl):
        printed_name = mdl.name or "AnonymousModel"
        out.write("// model name is: {0:s}\n".format(printed_name))

    def _vartype_name(self, vartype):
        # INTERNA: returns a printable string for a vartype
        return self.vartype_map.get(type(vartype), "unknown")

    def _print_var_containers(self, out, mdl):
        gensym_count = 1
        printed_header = False
        for ctn in mdl.iter_var_containers():
            if not printed_header:
                self._print_line_comment(out, "var contrainer section")
                printed_header = True
            vartype_name = self._vartype_name(ctn.vartype)
            varctn_name = ctn.name
            if not varctn_name:
                varctn_name = 'x%d' % gensym_count
                gensym_count += 1
            out.write("dvar {0} {1}{2};\n".format(vartype_name, varctn_name, ctn.dimension_string))

        if printed_header:
            self._newline(out)

    def _print_single_vars(self, out, mdl):
        printed_header = False
        for v in mdl.iter_variables():
            var_ctn = v.get_container()
            if var_ctn is not None:
                continue

            if not printed_header:
                self._print_line_comment(out, "single vars section")
                printed_header = True
            vartype_name = self._vartype_name(v.vartype)
            var_printname = self._var_name_map.get(v._index, "???")
            v_origin = v.origin()
            s_generated = ' # -- generated by: {0!s}'.format(v_origin) if v_origin is not None else ''
            out.write("dvar {0} {1};{2}\n".format(vartype_name, var_printname, s_generated))

        if printed_header:
            self._newline(out)

    def _print_objective(self, wrapper, model):
        wrapper.write(model.objective_sense.verb)
        wrapper.flush(print_newline=True)
        objexpr = model.objective_expr
        objlin = objexpr.get_linear_part()
        printed = self._print_lexpr(wrapper, self._num_printer, self._var_name_map, objlin,
                                    allow_empty=True,
                                    print_constant=False)
        if objexpr.is_quad_expr() and objexpr.has_quadratic_term():
            self._print_qexpr_obj(wrapper, self._num_printer, self._var_name_map,
                                  quad_expr=objexpr,
                                  force_initial_plus=printed,
                                  use_double=False)
            printed = True
        obj_offset = objexpr.get_constant()
        if obj_offset:
            if printed and obj_offset > 0:
                wrapper.write(u'+')
            wrapper.write(self._num_to_string(obj_offset))
        wrapper.write(';', separator=False)
        wrapper.flush()

    def _pprint_expr(self, wrapper, expr):
        q = 0
        if expr.is_quad_expr() and expr.has_quadratic_term():
            q = self._print_qexpr_iter(wrapper, self._num_printer, self._var_name_map, expr.iter_sorted_quads(),
                                       use_double=False)
        self._print_expr_iter(wrapper, self._num_printer, self._var_name_map, expr.iter_terms(),
                              constant=expr.get_constant(),  # yes, print the constant
                              allow_empty=q > 0,
                              force_first_plus=q > 0  # force  a '+' if quadratic section is non-empty
                              )

    def _print_binary_constraint(self, wrapper, ct):
        left_expr = ct.left_expr
        right_expr = ct.right_expr
        self._pprint_expr(wrapper, left_expr)

        wrapper.write(self.ct_symbol_map[ct.sense])

        self._pprint_expr(wrapper, right_expr)

    def _print_range_constraint(self, wrapper, rng):
        expr = rng.expr
        lb = rng.lb
        ub = rng.ub
        wrapper.write(self._num_to_string(lb))
        wrapper.write("<=")
        self._print_lexpr(wrapper, self._num_printer, self._var_name_map, expr, print_constant=True)
        wrapper.write("<=")
        wrapper.write(self._num_to_string(ub))

    def _print_logical_constraint(self, wrapper, logi_ct, logical_symbol):
        active = logi_ct.active_value
        linear_ct = logi_ct.linear_constraint
        indicator_varname = self._var_print_name(logi_ct.binary_var)
        wrapper.write(indicator_varname)
        if 0 == active:
            wrapper.write("== 0")
        wrapper.write(logical_symbol)
        wrapper.write('(')
        self._print_binary_constraint(wrapper, linear_ct)
        wrapper.write(');', separator=False)

    def _print_linear_constraints(self, wrapper, model):
        wrapper.begin_line()
        for ct in model.iter_binary_constraints():

            wrapper.begin_line()
            ctname = self.linearct_print_name(ct)
            if ctname:
                wrapper.set_indent('  ')  # two spaces
                wrapper.write(" %s:" % ctname)
                wrapper.flush()
            else:
                wrapper.begin_line(indented=True)

            self._print_binary_constraint(wrapper, ct)
            wrapper.write(';', separator=False)
            wrapper.set_indent(' ')
            wrapper.flush(print_newline=True, restart_from_empty_line=False)

    def _print_ranges(self, wrapper, model):
        wrapper.begin_line()
        for ct in model.iter_range_constraints():

            wrapper.begin_line()
            ctname = self.linearct_print_name(ct)
            if ctname:
                wrapper.set_indent(2 * ' ')
                wrapper.write(' %s:' % ctname)
                wrapper.flush()

            else:
                wrapper.begin_line(indented=True)

            self._print_range_constraint(wrapper, ct)
            wrapper.write(';', separator=False)
            wrapper.flush()
        wrapper.set_indent(' ')

    def _print_logicals(self, wrapper, model):
        for ct in model.iter_logical_constraints(include_implicits=True):
            ctname = self.logicalct_print_name(ct)
            if ctname:
                wrapper.set_indent('  ')
                wrapper.write(" %s:" % ctname)
                wrapper.flush()
            else:
                wrapper.begin_line(indented=True)
            symb = '<=>' if ct.is_equivalence() else '<='
            self._print_logical_constraint(wrapper, ct, logical_symbol=symb)
            wrapper.set_indent(' ')
            wrapper.flush(restart_from_empty_line=True)

    def _print_quadratic_cts(self, wrapper, model):
        for qct in model.iter_quadratic_constraints():

            ctname = self.qc_print_name(qct)
            if ctname:
                wrapper.set_indent('  ')
                wrapper.write(" %s:" % ctname)
                wrapper.flush()
            else:
                wrapper.begin_line(indented=True)

            self._print_binary_constraint(wrapper, qct)
            wrapper.write(';', separator=False)
            wrapper.set_indent(' ')
            wrapper.flush(restart_from_empty_line=True)

    def _print_kpis(self, out, wrapper, model):
        printed_section_header = False
        for kpi in model.iter_kpis():
            if kpi.is_decision_expression():
                if not printed_section_header:
                    self._newline(out)
                    self._print_line_comment(out, " KPI section")
                    printed_section_header = True

                kpi_expr = kpi.to_expr()
                kpi_typename = 'int' if kpi_expr.is_discrete() else 'float'
                wrapper.write('dexpr {0} {1}'.format(kpi_typename, self._translate_chars(kpi.name)))
                wrapper.write('=')
                if isinstance(kpi_expr, LinearExpr):
                    self._print_lexpr(wrapper, self._num_printer, self._var_name_map, kpi_expr, print_constant=True)
                elif isinstance(kpi_expr, Var):
                    wrapper.write(kpi_expr.name)
                wrapper.write(';', separator=False)
                wrapper.flush(restart_from_empty_line=True)
                # ---
        if printed_section_header:
            wrapper.newline()

    def _print_sos(self, wrapper, model):
        varname_map = self._var_name_map
        for sos_varset in model.iter_sos():
            sos_varname_joined = ', '.join(varname_map[v._index] for v in sos_varset.iter_variables())
            wrapper.write('sos{0}: {1}'.format(sos_varset.sos_type.value, sos_varname_joined))
        wrapper.flush()

    def print_model_to_stream(self, out, model):
        wrapper = _ExportWrapper(oss=out, indent_str=' ', line_width=78)
        self.prepare(model)
        # header
        self._print_signature(out)
        self._print_encoding(out)
        self._print_model_name(out, model)

        # var containers
        self._print_var_containers(out, model)
        self._print_single_vars(out, model)
        # KPI section
        self._print_kpis(out, wrapper, model)

        self._print_objective(wrapper, model)
        wrapper.write("\nsubject to {")
        wrapper.flush()
        self._print_linear_constraints(wrapper, model)
        self._print_ranges(wrapper, model)
        self._print_logicals(wrapper, model)
        self._print_quadratic_cts(wrapper, model)
        self._print_sos(wrapper, model)
        out.write("}\n")



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

from docplex.mp.engine import NoSolveEngine, ZeroSolveEngine
from docplex.mp.docloud_engine import DOcloudEngine
from docplex.mp.utils import is_string
from docplex.mp.context import has_credentials
from docplex.mp.error_handler import docplex_fatal


class EngineFactory(object):
    """ A factory class that manages creation of solver instances.
    """
    _default_engine_map = {"nosolve": NoSolveEngine,
                           "zero": ZeroSolveEngine,
                           "docloud": DOcloudEngine,
                           "cplexcloud": DOcloudEngine}

    def __init__(self, env=None):
        self._engine_types_by_agent = self._default_engine_map.copy()
        # no cplex engine type yet?
        if env is not None:
            self._resolve_cplex(env)

    def _get_engine_type_from_agent(self, agent, default_engine, default_engine_name):
        if agent is None:
            return default_engine
        elif is_string(agent):
            agent_key = agent.lower()
            engine_type = self._engine_types_by_agent.get(agent_key)
            if engine_type:
                return engine_type
            elif 'cplex' == agent_key:
                print('* warning: CPLEX DLL not found in path, using {0} instead'.format(default_engine_name))
                return self._engine_types_by_agent.get(default_engine_name)
            elif '.' in agent:
                # assuming a qualified name, e.g. com.ibm.docplex.quantum.QuantumEngine
                from docplex.mp.internal.mloader import import_class
                try:
                    agent_class = import_class(agent)
                    return agent_class
                except ValueError as ve:
                    print(
                        "Cannot load agent class {0}, expecting 'cplex', 'docloud' or valid class path, error: {1}".format(
                            agent, str(ve)))
                    raise ve
            else:
                docplex_fatal("Unexpected agent name: {0}, expecting 'cplex', 'docloud' or valid class path", agent)

        else:
            # try a class type
            try:
                # noinspection PyUnresolvedReferences
                from inspect import isclass
                if isclass(agent):
                    return agent
            except ImportError:
                if type(agent) == type:
                    return agent

            # agent cannot be mapped to any class.
            docplex_fatal("* unexpected agent: {0!r} -expecting 'cplex', 'docloud', class or class name", agent)

    def _is_cplex_resolved(self):
        return hasattr(self, "_cplex_engine_type")

    def _resolve_cplex(self, env):
        # INTERNAL
        if env is None:
            docplex_fatal("need an environment to resolve cplex, got None")
        if not self._is_cplex_resolved():
            if env.has_cplex:
                from docplex.mp.cplex_engine import CplexEngine

                self._cplex_engine_type = CplexEngine
                # noinspection PyTypeChecker
                self._engine_types_by_agent["cplex"] = CplexEngine
            else:
                self._cplex_engine_type = None

    def _ensure_cplex_resolved(self, env):
        if not self._is_cplex_resolved():
            self._resolve_cplex(env)
        assert self._is_cplex_resolved()

    def new_engine(self, agent, env, model, context=None):
        self._ensure_cplex_resolved(env)

        # compute a default engine and kwargs to use..
        kwargs = {}
        if self._cplex_engine_type:
            # CPLEX ahs been resolved and has a non-None type
            # default is CPLEX if we have it
            default_engine_type = self._cplex_engine_type
            default_engine_name = 'cplex'

        # elif has_credentials(context.solver.docloud):
        #     # default is docloud
        #     default_engine_type = DOcloudEngine
        #     default_engine_name = 'docloud'

        else:
            # no CPLEX, no credentials
            default_engine_type = NoSolveEngine
            default_engine_name = 'nosolve'

        if has_credentials(context.solver.docloud):
            kwargs['docloud_context'] = context.solver.docloud

        engine_type = self._get_engine_type_from_agent(agent=agent,
                                                       default_engine=default_engine_type,
                                                       default_engine_name=default_engine_name)
        assert engine_type is not None
        try:
            return engine_type(model, **kwargs)
        except TypeError:
            docplex_fatal("agent: {0!s} failed to create instance from model, kwargs.", agent)

    # noinspection PyMethodMayBeStatic
    def new_docloud_engine(self, model, **kwargs):
        # noinspection PyDeprecation
        return DOcloudEngine(model, **kwargs)


    def extend(self, new_agent, new_engine):
        # INTERNAL
        assert new_engine is not None
        self._engine_types_by_agent[new_agent] = new_engine



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2018
# --------------------------------------------------------------------------
'''This package provide simple logging facilities.
'''
import logging


class DocplexLogger(object):
    '''Simple logger interface.
    '''
    def log(self, level, msg):
        '''Logs a message with the specified level.
        '''
        raise NotImplementedError('subclasses must override log()!')

    def debug(self, msg):
        '''Logs a message with level DEBUG.
        '''
        self.log(logging.DEBUG, msg)

    def info(self, msg):
        '''Logs a message with level INFO.
        '''
        self.log(logging.INFO, msg)

    def warning(self, msg):
        '''Logs a message with level WARNING.
        '''
        self.log(logging.WARNING, msg)

    def error(self, msg):
        '''Logs a message with level ERROR.
        '''
        self.log(logging.ERROR, msg)

    def critical(self, msg):
        '''Logs a message with level CRITICAL.
        '''
        self.log(logging.CRITICAL, msg)


class LoggerToDocloud(DocplexLogger):
    '''This logger maps logs with python style logging levels to a docplexcloud
    logger expecting java style logging levels.
    '''
    def __init__(self, docloudlogger):
        super(LoggerToDocloud, self).__init__()
        self.docloudlogger = docloudlogger

    def log(self, level, msg):
        if level == logging.DEBUG:
            self.docloudlogger.fine(msg)
        elif level == logging.INFO:
            self.docloudlogger.info(msg)
        elif level == logging.WARNING:
            self.docloudlogger.warning(msg)
        elif level == logging.ERROR or level == logging.CRITICAL:
            self.docloudlogger.error(msg)
        else:
            raise ValueError('Supported logging levels are: DEBUG, INFO, WARNING, ERROR, CRITICAL. Provided = %s' % level)


class LoggerToFile(DocplexLogger):
    '''This logger logs log records with python style logging levels and just print them
    on the specified stream
    '''
    def __init__(self, file):
        super(LoggerToDocloud, self).__init__()
        self.file = file

    def log(self, level, msg):
        self.file.write('[%s] %s\n' % (level, msg))



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2016
# --------------------------------------------------------------------------

# gendoc: ignore


def as_df(what, **kwargs):
    '''
    Returns a `pandas.DataFrame` representation of an object.

    Attributes:
        what: The object to represent as an object.
        **kwargs: Additional parameters for the conversion.
    Returns:
        A `pandas.DataFrame` representation of an object or None if a
        representation could not be found.
    '''
    try:
        return what.__as_df__(**kwargs)
    except AttributeError:
        return None



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------


# gendoc: ignore
import logging
import os
import sched
import tempfile
import threading
import time
import sys

from six import PY2 as SIX_PY2
from six import itervalues
from itertools import chain, repeat

from docplex.mp.compat23 import Queue, izip


__int_types = {int}
__float_types = {float}
__numpy_ndslot_type = None
__numpy_matrix_type = None
__pandas_series_type = None
__pandas_dataframe_type = None
__spark_dataframe_type = None

try:
    # noinspection PyUnresolvedReferences
    type(long)  # @UndefinedVariable
    # long is indeed a type we are in Python2,
    __int_types.add(long)  # @UndefinedVariable
except NameError:  # pragma: no cover
    # long is not a type, do nothing
    pass  # pragma: no cover

try:
    import numpy

    _numpy_is_available = True

    __int_types.add(numpy.bool_)
    __int_types.add(numpy.bool)

    __int_types.add(numpy.int_)
    __int_types.add(numpy.intc)
    __int_types.add(numpy.intp)

    __int_types.add(numpy.int8)
    __int_types.add(numpy.int16)
    __int_types.add(numpy.int32)
    __int_types.add(numpy.int64)

    __int_types.add(numpy.uint8)
    __int_types.add(numpy.uint16)
    __int_types.add(numpy.uint32)
    __float_types.add(numpy.uint64)

    __float_types.add(numpy.float_)
    __float_types.add(numpy.float16)
    __float_types.add(numpy.float32)
    __float_types.add(numpy.float64)

    from numpy import ndarray, matrix

    __numpy_ndslot_type = ndarray
    __numpy_matrix_type = matrix
except ImportError:  # pragma: no cover
    _numpy_is_available = False  # pragma: no cover
    numpy_is_numeric = None
    numpy_is_integer = None

try:
    from pandas import Series, DataFrame

    __pandas_series_type = Series
    __pandas_dataframe_type = DataFrame
except ImportError:
    __pandas_series_type = None
    __pandas_dataframe_type = None

# 'findspark' must be executed if running in Windows environment, before importing Spark
try:
    import findspark  # @UnresolvedImport
    findspark.init()
except (ImportError, IndexError, ValueError):
    pass

try:
    import pyspark

    __spark_dataframe_type = pyspark.sql.dataframe.DataFrame
except ImportError:
    __spark_dataframe_type = None

__int_types = frozenset(__int_types)


def is_int(s):
    type_of_s = type(s)
    return type_of_s in __int_types or (_numpy_is_available and numpy_is_integer(type(s)))


__all_python_num_types = frozenset(__float_types.union(__int_types))

if _numpy_is_available:
    def numpy_is_numeric(t):
        # returns True if the specified type is numeric
        try:
            flag = numpy.issubdtype(t, numpy.number)
            global __all_python_num_types
            if flag:
                tmp = set(__all_python_num_types)
                tmp.add(type(t))
                __all_python_num_types = frozenset(tmp)
            return flag
        except TypeError:
            return False


    def numpy_is_integer(t):
        # returns True if the specified type is integer
        try:
            flag = numpy.issubdtype(t, numpy.integer)
            if flag is True:
                global __all_python_num_types
                global __int_types
                tmp = set(__all_python_num_types)
                tmp.add(type(t))
                __all_python_num_types = frozenset(tmp)

                tmp = set(__int_types)
                tmp.add(type(t))
                __int_types = frozenset(tmp)
            return flag
        except TypeError:
            return False


    def _is_numpy_ndslot(s):
        # returns True if the argument is a numpy number
        # wrapped in a fake ndarray
        # all the following conditions must be satisfied:
        # 1. numpy is present
        # 2. type is ndarray
        # 3. shape is () empty tuple
        # 4. wrapped type in ndarray is numeric.
        try:
            retval = is_numpy_ndarray(s) and s.shape == () and \
                     (s.dtype.type in __all_python_num_types or numpy_is_numeric(s.dtype))
            return retval
        except AttributeError:  # if s is not a numpy type, s.dtype triggers this
            return False
else:
    def numpy_is_numeric(t):
        return False


    def numpy_is_integer(t):
        return False


    def _is_numpy_ndslot(s):
        return False


def is_number(s):
    type_of_s = type(s)
    return type_of_s in __all_python_num_types or numpy_is_numeric(type_of_s) or _is_numpy_ndslot(s)


# def is_number2(s):
#     if isinstance(s, __all_num_types_tuple):
#         return True
#     elif _numpy_is_available:
#         type_of_s = type(s)
#         if (numpy_is_numeric(type_of_s) or _is_numpy_ndslot(s)):
#             return True
#     else:
#         return False

def is_pandas_series(s):
    return __pandas_series_type is not None and type(s) is __pandas_series_type


def is_pandas_dataframe(s):
    return __pandas_dataframe_type and isinstance(s, __pandas_dataframe_type)


def is_numpy_ndarray(s):
    return __numpy_ndslot_type and type(s) is __numpy_ndslot_type


def is_numpy_matrix(s):
    return __numpy_matrix_type and type(s) is __numpy_matrix_type


def is_spark_dataframe(s):
    return __spark_dataframe_type and isinstance(s, __spark_dataframe_type)


string_types = {str}
if SIX_PY2:
    string_types.add(unicode)  # @UndefinedVariable
string_types = frozenset(string_types)


def is_string(e):
    return type(e) in string_types

def has_len(e):
    try:
        len(e)
        return True
    except TypeError:
        return False


def is_indexable(e):
    ''' Returns true if it is indexable
    '''
    return hasattr(e, "__getitem__")


def is_iterable(e, accept_string=True):
    ''' Returns true if we can extract an iterator from it
    '''
    try:
        iter(e)
        return accept_string or not is_string(e)
    except TypeError:
        return False


def is_iterator(e):
    ''' Returns true if e is its own iterator.
    '''
    try:
        # some numpy iterators just fail on == but are ok with "is"
        return e is iter(e)
    except TypeError:
        return False


def is_function(e):
    import platform
    if platform.python_version() >= '3.7':
        from collections.abc import Callable
    else:
        from collections import Callable  # @UnresolvedImport
    return isinstance(e, Callable)


def _to_list(arg):
    # INTERNAL:
    # 1. checks the argument is either a sequence or iterator,;
    # if sequence, returns the sequence, else converts to a list by exhsuating the iterator
    # BEWARE of the infinite generator!
    if is_iterator(arg):
        return list(arg)
    elif is_iterable(arg):
        return arg
    else:
        # an atom: wrap it into a list
        return list(arg)


def _build_ordered_sequence_types():
    if __pandas_series_type and __numpy_ndslot_type:
        return (list, tuple, __pandas_series_type, __numpy_ndslot_type)
    elif __pandas_series_type:
        return (list, tuple, __pandas_series_type)
    elif __numpy_ndslot_type:
        return (list, tuple, __numpy_ndslot_type)
    else:
        return (list, tuple)

def is_ordered_sequence1(arg, type_tuple=_build_ordered_sequence_types()):
    return isinstance(arg, type_tuple)

def is_ordered_sequence2(arg):
    if isinstance(arg, (dict, str)):
        return False

    try:
        # try it
        l = len(arg)
        if l:
            arg[0]
        return True
    except:
        return False

is_ordered_sequence = is_ordered_sequence1


class DOcplexException(Exception):
    """ Base class for modeling exceptions 
    """
    DEFAULT_MSG = 'CplexPythonModeling exception raised'

    def __init__(self, msg, *args):
        Exception.__init__(self, msg)
        self.__msg = msg or self.__class__.DEFAULT_MSG
        self.__edited_message = None
        self.__args = args
        self._resolve_message()

    def _resolve_message(self):
        self.__edited_message = None
        msg = self.__msg
        if self.__args:
            if '%' in msg:
                self.__edited_message = msg % self.__args
            elif '{' in msg:
                self.__edited_message = msg.format(*self.__args)

    @property
    def message(self):
        return self.__edited_message or self.__msg


class DOcplexLimitsExceeded(DOcplexException):
    cplexce_limits_exceeded_msg = "**** Promotional version. Problem size limits exceeded, CPLEX code=1016"

    def __init__(self):
        DOcplexException.__init__(self, self.cplexce_limits_exceeded_msg)


class DocplexQuadToLinearException(DOcplexException):

    def __init__(self, q):
        msg = "Quadratic expression [{0!s}] cannot be converted to a linear expression".format(q)
        DOcplexException.__init__(self, msg)


class DocplexLinearRelaxationError(DOcplexException):

    def __init__(self, obj, cause=None):
        self.object = obj
        self.cause = cause
        msg = "Modeling object: {0!s} cannot be relaxed".format(obj)
        DOcplexException.__init__(self, msg)

def normalize_basename(s, force_lowercase=True, maxlen=255):
    """Replaces some characters from s with a translation table:
    
     trans_table = {" ": "_",
                   "/": "_slash_",
                   "\\": "_backslash_",
                   "?": "_question_",
                   "%": "_percent_",
                   "*": "_asterisk_",
                   ":": "_colon_",
                   "|": "_bar_",
                   '"': "_quote_",
                   "<": "_lt_",
                   ">": "_gt_",
                   "&": "_amp_"}
    
    then if the generated name is longer than maxlen, the name is truncated
    to maxlen and the hash of the name modulo 0xffffffff is appended.
    """
    # replace all whietspaces by _
    l = s.lower() if force_lowercase else s
    # table = mktrans(" ", "_")
    # return l.translate(table)
    trans_table = {" ": "_",
                   "/": "_slash_",
                   "\\": "_backslash_",
                   "?": "_question_",
                   "%": "_percent_",
                   "*": "_asterisk_",
                   ":": "_colon_",
                   "|": "_bar_",
                   '"': "_quote_",
                   "<": "_lt_",
                   ">": "_gt_",
                   "&": "_amp_"}

    n = ("".join([trans_table.get(x, x) for x in l]))

    if len(n) > maxlen - 8:
        h = format(hash(n) & 0xffffffff, "08x")
        n = n[:maxlen-8] + "_"+ h
    return n

def fix_whitespace(s, fix='_'):
    # replaces whitespaces by a character, default is '_'
    return s.replace(' ', fix)


def make_output_path2(actual_name, extension, basename_fmt, path=None):
    # INTERNAL
    raw_basename = resolve_pattern(basename_fmt, actual_name) if basename_fmt else actual_name
    # fix whitespaces
    if raw_basename.find(" ") > 0:
        actual_basename = raw_basename.replace(" ", "_")
    else:
        actual_basename = raw_basename

    output_dir = path or tempfile.gettempdir()
    # fix extension
    if not actual_basename.endswith(extension):
        actual_basename = actual_basename + extension
    return os.path.join(output_dir, actual_basename)


def make_path(error_handler, basename, extension, output_dir=None, name_transformer=None):
    if output_dir is None:
        output_dir = tempfile.gettempdir()
    elif not os.path.exists(output_dir):
        if not os.makedirs(output_dir):
            error_handler.error("directory not found and not created: {0:s}", output_dir)
            return None

    norm_name = normalize_basename(basename)
    basename = norm_name if not name_transformer else name_transformer % norm_name
    filename = basename + extension
    full_path = '/'.join([output_dir, filename])
    return full_path


def generate_constant(the_constant, count_max):
    if count_max is None:
        count_max = sys.maxsize
    loop_counter = 0
    while loop_counter <= count_max:
        yield the_constant
        loop_counter += 1


def iter_emptyset():
    return iter([])

def iter_one(obj):
    yield obj



def resolve_pattern(pattern, args):
    """
    returns a string in which slots have been resolved with args, if the string has slots anyway,
    else returns the strng itself (no copy, should we??)
    :param pattern:
    :param args:
    :return:
    """
    if args is None or len(args) == 0:
        return pattern
    elif pattern.find('%') >= 0:
        return pattern % args
    elif pattern.find("{") >= 0:
        # star magic does not work for single args
        return pattern.format(*args)
    else:
        # fixed pattern, no placeholders
        return pattern


def str_maxed(arg, maxlen):
    """ Returns a (possibly) truncated string representation of arg

    If maxlen is positive (or null), returns str(arg) up to maxlen chars.

    :param arg:
    :param maxlen:
    :return:
    """
    s = str(arg)
    if maxlen <= 0 or len(s) <= maxlen:
        return s
    else:
        return "%s.." % s[:maxlen]


try:
    import scipy.sparse as sp
except ImportError:
    sp = None

def is_scipy_sparse(m):
    return sp and sp.issparse(m)


def compute_is_index(seq, obj):
    # assume obj is iterable multiple times
    for i, elt in enumerate(seq):
        if elt is obj:  # use 'is' to identify obj
            return i
    else:
        return None


DOCPLEX_CONSOLE_HANDLER = None


def get_logger(name, verbose=False):
    logging_level = logging.WARNING
    if verbose:
        logging_level = logging.DEBUG

    logger = logging.getLogger(name)
    logger.setLevel(logging_level)

    global DOCPLEX_CONSOLE_HANDLER

    if DOCPLEX_CONSOLE_HANDLER is None:
        DOCPLEX_CONSOLE_HANDLER = logging.StreamHandler()
        DOCPLEX_CONSOLE_HANDLER.setLevel(logging_level)

    if DOCPLEX_CONSOLE_HANDLER not in logger.handlers:
        logger.addHandler(DOCPLEX_CONSOLE_HANDLER)

    return logger


def open_universal_newline(filename, mode):
    """Opens a file in universal new line mode, in a python 2 and python 3
    compatible way.
    """
    try:
        # try the python 3 syntax
        return open(filename, mode=mode, newline=None)
    except TypeError as te:
        if "'newline'" in str(te):
            # so open does not have a newline parameter -> python 2, use "U"
            # mode
            return open(filename, mode=mode + "U")
        else:
            # for other errors, just raise them
            raise


class CyclicLoop(object):
    """ A cyclic loop executes actions at specified intervals, until
    ``stop()`` is called.

    This loop is based on sched.scheduler

    Attributes:
        stopped: True if the loop is stopped.
    """

    class Task(object):
        """This class stores information needed to manage tasks.

        Attributes:
            id: The id of the task (automatically generated)
            interval: The interval on which that task is called
            action: The action function to call at ``interval``
            argument: The arguments for the action function
        """
        id = 0
        idgen_lock = threading.Lock()

        def __init__(self, interval, priority, action, argument=()):
            self.interval = interval
            self.priority = priority
            self.action = action
            self.argument = argument
            with self.idgen_lock:
                self.id = CyclicLoop.Task.id
                CyclicLoop.Task.id += 1

    def __init__(self):
        """Initialize a new empty CyclicLoop
        """
        self.stop_lock = threading.Lock()
        self.stopped = False
        self.scheduler = sched.scheduler(time.time, time.sleep)
        # maps task id -> ev
        self.events_by_id = {}
        self.tasks_by_id = {}  # task id -> task

    def enter(self, interval, priority, action, argument=()):
        """Schedule a new event.

        Works like sched.scheduler.enter(), but instead of a ``delay``, the
        first argument is the ``interval`` the action must be performed.
        """
        with self.stop_lock:
            if not self.stopped:
                task = CyclicLoop.Task(interval, priority, action, argument)
                self.tasks_by_id[task.id] = task
                self._queue(task)

    def _queue(self, task):
        ev = self.scheduler.enter(task.interval, task.priority,
                                  lambda a: self._process_task(a), (task.id,))
        self.events_by_id[task.id] = ev

    def _process_task(self, task_id):
        task = self.tasks_by_id[task_id]
        task.action(*task.argument)
        del self.events_by_id[task_id]
        # do not reschedule if we are shutting down
        with self.stop_lock:
            if not self.stopped:
                self._queue(task)

    def start(self):
        """Starts the loop. The loop stops only when ``stop()`` is called.
        """
        self.scheduler.run()

    def stop(self):
        """Stops the Loop.

        When the loop is stopped, its ``stopped`` attribute is set immediately,
        then all tasks in the scheduler are canceled.
        """
        with self.stop_lock:
            self.stopped = True
            for ev in itervalues(self.events_by_id):
                try:
                    self.scheduler.cancel(ev)
                except ValueError:
                    # if stop() is called from an event, the event has already
                    # been triggered and poped'
                    pass


class ClosableQueue(Queue):
    LAST = object()

    def close(self):
        self.put(self.LAST)

    def __iter__(self):
        while True:
            item = self.get()
            try:
                if item is self.LAST:
                    return
                yield item
            finally:
                self.task_done()


class ThreadedCyclicLoop(object):
    """ A cyclic loop executes actions at specified intervals, until
    ``stop()`` is called.

    This loop is based on threads.

    Attributes:
        stopped: True if the loop is stopped.
    """

    class Task(threading.Thread):
        """
        Attributes:
            id: The id of the task (automatically generated)
            interval: The interval on which that task is called
            action: The action function to call at ``interval``
            argument: The arguments for the action function
        """

        def __init__(self, loop, interval, priority, action, argument=()):
            super(ThreadedCyclicLoop.Task, self).__init__()
            self.loop = loop
            self.interval = interval
            self.priority = priority
            self.action = action
            self.argument = argument
            self.stopped = False

        def run(self):
            while not self.stopped:
                # instead of one big sleep, do some smaller sleeps so that
                # we can stop the thread with smaller granularity
                for _ in range(self.interval):
                    time.sleep(1)
                    if self.stopped:
                        break
                if not self.stopped:
                    self.perform()

        def stop(self):
            self.stopped = True

        def perform(self):
            self.action(*self.argument)

    def __init__(self):
        """Initialize a new empty ThreadedCyclicLoop
        """
        self.stop_lock = threading.Lock()
        self.stopped = False
        self.threads = set()
        self.event_queue = ClosableQueue()

    def enter(self, interval, priority, action, argument=()):
        """Schedule a new event.

        Works like sched.scheduler.enter(), but instead of a ``delay``, the
        first argument is the ``interval`` the action must be performed.
        """
        with self.stop_lock:
            if not self.stopped:
                task = ThreadedCyclicLoop.Task(self, interval, priority,
                                               action, argument)
                self.threads.add(task)

    def start(self, mt_worker=None, mt_arg=()):
        """Starts the loop. The loop stops only when ``stop()`` is called.
        """
        for t in self.threads:
            t.start()
        if mt_worker:
            while not self.stopped:
                for task in self.event_queue:
                    mt_worker(*((task,) + mt_arg))
        for t in self.threads:
            t.join()

    def stop(self):
        """Stops the Loop.

        When the loop is stopped, its ``stopped`` attribute is set immediately,
        then all tasks in the scheduler are canceled.
        """
        with self.stop_lock:
            self.stopped = True
        self.event_queue.close()
        for t in self.threads:
            t.stop()


class _SymbolGenerator(object):
    def __init__(self, pattern, offset=1):
        ''' Initialize the counter and the pattern.
            Fixes the pattern by suffixing '%d' if necessary.
        '''
        self._pattern = pattern
        self._offset = offset
        self._set_pattern(pattern)

    def _set_pattern(self, pattern):
        if pattern.endswith('%d'):
            self._pattern = pattern
        else:
            self._pattern = pattern + '%d'

    def new_index_symbol(self, index):
        return self._pattern % (index + self._offset)

    def new_obj_symbol(self, obj):
        return self.new_index_symbol(obj.index)


class _AutomaticSymbolGenerator(_SymbolGenerator):

    def __init__(self, pattern, offset=1):
        ''' Initialize the counter and the pattern.
            Fixes the pattern by suffixing '%d' if necessary.
        '''
        _SymbolGenerator.__init__(self, pattern, offset)
        self._last_index = -1

    def reset(self):
        self._last_index = -1

    def notify_new_index(self, new_index):
        # INTERNAL
        if new_index > self._last_index:
            self._last_index = new_index

    def new_symbol(self):
        """
        Generates and returns a new symbol.
        Guess a new (yet) unallocated index, then use the pattern.
        Note that we use the offset of 1 to generate the name so x1 has index 0, x3 has index 2, etc.
        :return: A symbol string, suposedly not yet allocated.
        """
        guessed_index = self._last_index + 1
        coined_symbol = _SymbolGenerator.new_index_symbol(self, guessed_index)
        self._last_index = guessed_index
        return coined_symbol

    def __iter__(self):
        while True:
            yield self.new_symbol()

class _IndexScope(object):
    # INTERNAL: full scope of indices.

    def __init__(self, obj_iter, qualifier, cplex_scope=None, offset=1):
        self._offset = offset
        self._obj_iter = obj_iter
        self._index_map = None
        self.cplex_scope = cplex_scope
        self.qualifier = qualifier
        self._last_index = -1

    def _make_index_map(self):
        return {m.get_index(): m for m in self._obj_iter()}

    def number_of_objects(self):
        idxmap = self._index_map
        return 0 if idxmap is None else len(idxmap)

    @property
    def iter(self):
        return self._obj_iter()

    def __call__(self, idx):
        return self._object_by_index(idx)

    def get_object_by_index(self, idx, checker=None):
        if checker:
            checker.typecheck_valid_index(idx)
        return self._object_by_index(idx)

    def _object_by_index(self, idx, do_raise=False):
        if self._index_map is None:
            self._index_map = self._make_index_map()
        # do not raise when not found, return None.
        if do_raise:
            return self._index_map[idx]
        else:
            # returns None if idx not in map
            return self._index_map.get(idx)

    def reset(self):
        self._index_map = None
        self._last_index = -1

    def notify_obj_index(self, obj, index):
        if self._index_map is not None:
            self._index_map[index] = obj

    def notify_obj_indices(self, objs, indices):
        # take the last one??
        if indices:
            idxmap = self._index_map
            if idxmap is not None:
                for obj, idx in izip(objs, indices):
                    idxmap[idx] = obj

    def update_indices(self):
        if self._index_map is not None:
            self._index_map = self._make_index_map()

    def reindex_all(self, indexer):
        for ix, obj in enumerate(self._obj_iter()):
            obj.set_index(ix)

    def notify_deleted(self, oldidx):
        if self._index_map:
            try:
                del self._index_map[oldidx]
            except KeyError:
                pass


def apply_thread_limitations(context, solver_context):
    # --- limit threads if needed
    parameters = context._get_raw_cplex_parameters()
    if getattr(solver_context, 'max_threads', None) is not None:
        parameters = context.cplex_parameters
        if parameters.threads.get() == 0:
            max_threads = solver_context.max_threads
        else:
            max_threads = min(solver_context.max_threads, parameters.threads.get())
        # we don't want to duplicate parameters unnecessary
        if max_threads != parameters.threads.get():
            new_parameters = parameters.copy()
            # change actual #threads
            new_parameters.threads = max_threads
            out_stream = solver_context.log_output_as_stream
            if out_stream:
                out_stream.write(
                    "WARNING: Number of workers has been reduced to %s to comply with platform limitations.\n" % max_threads)
            # --- here we copy the initial parameters ---
            return new_parameters

    return parameters


def compute_overwrite_nb_threads(parameters, solver_context):
    # returns a corrected value for threads, if necessary, else None
    if getattr(solver_context, 'max_threads', None) is not None:
        solve_threads = solver_context.max_threads
        param_threads = parameters.threads.get()
        if 0 == param_threads:
            # 0 for cplex meeans any number
            max_threads = solve_threads
        else:
            max_threads = min(solve_threads, param_threads)
        if max_threads != param_threads:
            return max_threads
    # None means no overwrite
    return None


def is_almost_equal(x1, x2, reltol, abstol=0):
    # returns true if x2 equals x1 w.r.t a miax of an absolute tolerance and a relative tolerance
    prec = max(abstol, reltol* abs(x1))
    return abs(x2-x1) <= prec


class OutputStreamAdapter:
    # With this class, we kind of automatically handle binary/non binary output streams
    # it automatically perform encoding of strings when needed,
    # and if the stream is a String stream, strings are just written without conversion
    def __init__(self, stream, encoding='utf-8'):
        self.stream = stream
        self.stream_is_binary = False
        if hasattr(stream, 'mode') and 'b' in stream.mode:
            self.stream_is_binary = True
        from io import TextIOBase
        if not isinstance(stream, TextIOBase):
            self.stream_is_binary = True

        self.encoding = encoding

    def write(self, s):
        # s is supposed to be a string
        output_s = s
        if self.stream_is_binary:
            output_s = s.encode(self.encoding)
        self.stream.write(output_s)

def izip2_filled(it1, it2, **kwds):
    # izip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
    fillvalue = kwds.get('fillvalue')

    class _ZipExhausted(Exception):
        pass

    class _sentinel():
        def __iter__(self):
            raise _ZipExhausted

    fillers = repeat(fillvalue)
    iterators = [chain(it1, _sentinel()), chain(it2, fillers)]
    try:
        while iterators:
            tpl = tuple(map(next, iterators))
            #print(tpl)
            yield tpl
    except _ZipExhausted:
        pass


class MultiObjective(object):
    # a wrapper class to hold all data for multi-objective

    def __init__(self, exprs, priorities, weights=None, abstols=None, reltols=None, names=None):
        self.exprs = exprs
        self.priorities = priorities
        self.weights = weights
        self.abstols = abstols
        self.reltols = reltols
        self.names = names

    @classmethod
    def new_empty(cls):
        return MultiObjective([], [])

    def empty(self):
        return not self.exprs

    @property
    def number_of_objectives(self):
        nb_exprs = len(self.exprs)
        return 0 if nb_exprs <= 1 else nb_exprs

    def clear(self):
        self.exprs = []
        self.priorities = []
        self.weights = []
        self.abstols = None
        self.reltols = None
        self.names = None

    def convert_pseudo_sequence(self, seq, size, default=None, pred=is_number):
        if pred and pred(seq):
            return [seq] * size
        elif is_indexable(seq):
            return seq
        elif seq is None:
            return [default] * size
        else:
            raise TypeError

    def iter_exprs(self):
        multi_obj_exprs = self.exprs
        if len(multi_obj_exprs) >= 2:
            return iter(multi_obj_exprs)
        else:
            return iter([])

    @staticmethod
    def as_optional_sequence(opt_seq, size):
        # converts the argument to None or a list,
        # special case is plain number, converted to a list of expected size.
        if is_number(opt_seq):
            return [opt_seq] * size
        elif opt_seq is None or is_indexable(opt_seq):
            # expecting either None or a list.
            return opt_seq
        else:
            raise TypeError('unexpected optional number sequence: {0!r}'.format(opt_seq))

    def abstols_as_sequence(self, size):
        return self.as_optional_sequence(self.abstols, size)

    def reltols_as_sequence(self, size):
        return self.as_optional_sequence(self.reltols, size)

    def update(self, new_exprs, new_prios, new_weights, new_abstols, new_reltols, new_names):
        self.exprs = new_exprs
        self.priorities = new_prios
        self.weights = new_weights
        self.abstols = new_abstols
        self.reltols = new_reltols
        self.names = new_names

    def itertuples(self):
        # iterates on all components
        # returns tuples of the form (expr, prio, weight, abstol, reltol, name)
        if self.empty():
            return iter([])

        # not empty
        nb_objs = len(self.exprs)
        assert nb_objs >= 2
        # default weight
        lw = self.convert_pseudo_sequence(self.weights, size=nb_objs, default=1)
        la = self.convert_pseudo_sequence(self.abstols, size=nb_objs, default=0)
        lr = self.convert_pseudo_sequence(self.reltols, size=nb_objs, default=0)
        ln = self.convert_pseudo_sequence(self.names, size=nb_objs, default=None, pred=None)

        return iter(zip(self.exprs, self.priorities, lw, la, lr, ln))


def resolve_caller_as_string(caller, sep=': '):
    # resolve caller as a string
    if caller is None:
        return ""
    else:
        try:
            return "%s%s" % (caller(), sep)
        except TypeError:
            return "%s%s" % (caller, sep)

    


# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------


from docplex.mp.solution import SolveSolution
from docplex.mp.sdetails import SolveDetails
from docplex.mp.utils import DOcplexException
from docplex.mp.error_handler import docplex_fatal
from docplex.util.status import JobSolveStatus
from docplex.mp.constants import CplexScope


# gendoc: ignore


class IEngine(object):

    def only_cplex(self, mname):
        raise TypeError("{0} requires CPLEX - not available.".format(mname))  # pragma: no cover

    def register_callback(self, cb):
        raise NotImplementedError  # pragma: no cover

    def connect_progress_listeners(self, listeners, model):
        raise NotImplementedError  # pragma: no cover

    def solve(self, mdl, parameters, **kwargs):
        ''' Redefine this method for the real solve.
            Returns a solution object or None.
        '''
        raise NotImplementedError  # pragma: no cover

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):
        """
        Runs feasopt-like algorithm with a set of relaxable cts with preferences

        :param mdl: The model for which relaxation is performed.
        :param relaxable_groups:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):
        """
        Runs conflict-refiner algorithm with an optional set constraints groups with preferences

        :param mdl: The model for which conflict refinement is performed.
        :param preferences: an optional dictionary defining constraints preferences.
        :param groups: an optional list of 'docplex.mp.conflict_refiner.ConstraintsGroup'.
        :param parameters:
        :return: A list of "TConflictConstraint" namedtuples, each tuple corresponding to a constraint that is
            involved in the conflict.
            The fields of the "TConflictConstraint" namedtuple are:
                - the name of the constraint or None if the constraint corresponds to a variable lower or upper bound
                - a reference to the constraint or to a wrapper representing a Var upper or lower bound
                - an :enum:'docplex.mp.constants.ConflictStatus' object that indicates the
                conflict status type (Excluded, Possible_member, Member...)
            This list is empty if no conflict is found by the conflict refiner.
        """
        raise NotImplementedError  # pragma: no cover

    def populate(self):
        raise NotImplementedError  # pragma: no cover

    def get_solve_status(self):
        """  Return a DOcplexcloud-style solve status.

        Possible enums are in docloud/status.py
        Default is UNKNOWN at this stage. Redefined for CPLEX and DOcplexcloud engines.
        """
        return JobSolveStatus.UNKNOWN  # pragma: no cover

    def get_cplex(self):
        """
        Returns the underlying CPLEX, if any. May raise an exception if not applicable.
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def has_cplex(self):  # pragma: no cover
        try:
            return self.get_cplex() is not None
        except DOcplexException:
            # some engine may raise an exception when accessing a cplex
            return False

    def set_parameter(self, parameter, value):
        """ Changes the parameter value.
        :param parameter:
        :param value:
        """
        raise NotImplementedError  # pragma: no cover

    def get_parameter(self, parameter):
        raise NotImplementedError  # pragma: no cover

    def get_solve_details(self):
        raise NotImplementedError  # pragma: no cover

    def clean_before_solve(self):
        raise NotImplementedError  # pragma: no cover

    def supports_logical_constraints(self):
        raise NotImplementedError  # pragma: no cover

    def solved_as_mip(self):
        return False

    @property
    def name(self):
        raise NotImplementedError  # pragma: no cover

    def get_infinity(self):
        raise NotImplementedError  # pragma: no cover

    def create_one_variable(self, vartype, lb, ub, name):
        raise NotImplementedError  # pragma: no cover

    def create_variables(self, nb_vars, vartype, lb, ub, name):
        raise NotImplementedError  # pragma: no cover

    def create_multitype_variables(self, keys, vartypes, lbs, ubs, names):
        raise NotImplementedError  # pragma: no cover

    def create_linear_constraint(self, binaryct):
        raise NotImplementedError  # pragma: no cover

    def create_block_linear_constraints(self, ct_seq):
        raise NotImplementedError  # pragma: no cover

    def create_range_constraint(self, rangect):
        raise NotImplementedError  # pragma: no cover

    def create_logical_constraint(self, logct, is_equivalence):
        raise NotImplementedError  # pragma: no cover

    def create_batch_logical_constraints(self, logcts, is_equivalence):
        # the default is to iterate and append.
        return [self.create_logical_constraint(logc, is_equivalence) for logc in logcts]

    def create_quadratic_constraint(self, qct):
        raise NotImplementedError  # pragma: no cover

    def create_pwl_constraint(self, pwl_ct):
        raise NotImplementedError  # pragma: no cover

    def remove_constraint(self, ct):
        raise NotImplementedError  # pragma: no cover

    def remove_constraints(self, cts):
        raise NotImplementedError  # pragma: no cover

    def set_objective_sense(self, sense):
        raise NotImplementedError  # pragma: no cover

    def set_objective_expr(self, new_objexpr, old_objexpr):
        raise NotImplementedError  # pragma: no cover

    def set_multi_objective_exprs(self, new_multiobjexprs, old_multiobjexprs,
                                  priorities=None, weights=None, abstols=None, reltols=None, objnames=None):
        raise NotImplementedError  # pragma: no cover

    def end(self):
        raise NotImplementedError  # pragma: no cover

    def set_streams(self, out):
        raise NotImplementedError  # pragma: no cover

    def set_var_lb(self, var, lb):
        raise NotImplementedError  # pragma: no cover

    def set_var_ub(self, var, ub):
        raise NotImplementedError  # pragma: no cover

    def rename_var(self, var, new_name):
        raise NotImplementedError  # pragma: no cover

    def change_var_types(self, dvars, new_types):
        raise NotImplementedError  # pragma: no cover

    def update_objective_epxr(self, expr, event, *args):
        raise NotImplemented  # pragma: no cover

    def update_constraint(self, ct, event, *args):
        raise NotImplementedError  # pragma: no cover

    def check_var_indices(self, dvars):
        raise NotImplementedError  # pragma: no cover

    def check_constraint_indices(self, cts, ctscope):
        raise NotImplementedError  # pragma: no cover

    def create_sos(self, sos):
        raise NotImplementedError  # pragma: no cover

    def clear_all_sos(self):
        raise NotImplementedError  # pragma: no cover

    def get_basis(self, mdl):
        raise NotImplementedError  # pragma: no cover

    def set_lp_start(self, var_stats, ct_stats):  # pragma: no cover
        raise NotImplementedError

    def export(self, out, fmt):  # pragma: no cover
        raise NotImplementedError


# noinspection PyAbstractClass
class MinimalEngine(IEngine):
    # define as most methods as possible with reasonable defaults.

    def export(self, out, fmt):
        self.only_cplex("export to %s" % fmt.name)

    def end(self):
        pass  # pragma: no cover

    def set_streams(self, out):
        pass  # pragma: no cover

    def register_callback(self, cb):
        # no callbacks
        pass   # pragma: no cover

    def connect_progress_listeners(self, listeners, model):
        # no listeners
        self.only_cplex(mname="connect_progress_listeners")   # pragma: no cover

    def set_parameter(self, parameter, value):
        #
        pass   # pragma: no cover

    def get_parameter(self, parameter):
        # engine value of a parameter is its own
        return parameter.get()   # pragma: no cover

    def clean_before_solve(self):
        pass   # pragma: no cover

    def get_infinity(self):
        return 1e+20   # pragma: no cover

    def create_one_variable(self, vartype, lb, ub, name):
        # define create one var in terms of create batch vars
        xs = self.create_variables(1, vartype, [lb], [ub], [name])
        return xs[0]   # pragma: no cover

    def set_var_lb(self, var, lb):
        pass   # pragma: no cover

    def set_var_ub(self, var, ub):
        pass   # pragma: no cover

    def rename_var(self, var, new_name):
        pass  # pragma: no cover

    def rename_linear_constraint(self, linct, new_name):
        pass  # pragma: no cover

    def check_var_indices(self, dvars):
        pass  # pragma: no cover

    def check_constraint_indices(self, cts, ctscope):
        pass  # pragma: no cover

    def remove_constraints(self, cts):
        pass   # pragma: no cover

    def update_constraint(self, ct, event, *args):
        self.only_cplex(mname="update_constraint")

    def update_extra_constraint(self, lct, qualifier, *args):
        self.only_cplex(mname="update_extra_constraint")

    def remove_constraint(self, ct):
        pass   # pragma: no cover

    def create_sos(self, sos):   # pragma: no cover
        self.only_cplex("create SOS set")

    def clear_all_sos(self):
        pass   # pragma: no cover

    def create_linear_constraint(self, ct):
        return self.create_block_linear_constraints([ct])[0]   # pragma: no cover

    def get_solve_details(self):
        from docplex.mp.sdetails import SolveDetails
        return SolveDetails()   # pragma: no cover

    def get_basis(self, mdl):
        self.only_cplex("get_basis")   # pragma: no cover

    def set_lp_start(self, var_stats, ct_stats):
        self.only_cplex("set_lp_start")   # pragma: no cover

    def supports_logical_constraints(self):
        return False   # pragma: no cover

    def change_var_types(self, dvars, new_types):
        self.only_cplex(mname="change_var_type")   # pragma: no cover

    def get_cplex(self):
        docplex_fatal("No cplex is available.")   # pragma: no cover

    def create_batch_logical_constraints(self, logcts, is_equivalence):
        self.only_cplex(mname="create_batch_logical_constraints")   # pragma: no cover

    def create_logical_constraint(self, logct, is_equivalence):
        self.only_cplex(mname="create_logical_constraint")   # pragma: no cover

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):
        self.only_cplex(mname="refine_conflict")   # pragma: no cover

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):
        self.only_cplex(mname="solve_relaxed")  # pragma: no cover

    def populate(self):
        self.only_cplex(mname="populate")  # pragma: no cover

    def create_pwl_constraint(self, pwl_ct):
        self.only_cplex(mname="create_quadratic_constraint")  # pragma: no cover

    def create_quadratic_constraint(self, qct):
        self.only_cplex(mname="create_quadratic_constraint")  # pragma: no cover

    def set_multi_objective_exprs(self, new_multiobjexprs, old_multiobjexprs,
                                  priorities=None, weights=None, abstols=None, reltols=None, objnames=None):
        self.only_cplex(mname="set_multi_objective_exprs")  # pragma: no cover



# noinspection PyAbstractClass,PyMethodMayBeStatic
class DummyEngine(IEngine):
    def export(self, out, fmt):
        self.only_cplex("export to %s" % fmt.name)

    def create_range_constraint(self, rangect):
        return -1  # pragma: no cover

    def create_logical_constraint(self, logct, is_equivalence):
        return -1  # pragma: no cover

    def create_quadratic_constraint(self, qct):
        return -1  # pragma: no cover

    def create_pwl_constraint(self, pwl_ct):
        return -1  # pragma: no cover

    def set_streams(self, out):
        pass  # pragma: no cover

    def get_infinity(self):
        return 1e+20  # pragma: no cover

    def create_one_variable(self, vartype, lb, ub, name):
        return -1  # pragma: no cover

    def create_variables(self, nb_vars, vartype, lb, ub, name):
        return [-1] * nb_vars  # pragma: no cover

    def create_multitype_variables(self, size, vartypes, lbs, ubs, names):
        return [-1] * size

    def set_var_lb(self, var, lb):
        pass

    def set_var_ub(self, var, ub):
        pass

    def rename_var(self, var, new_name):
        pass  # nothing to do, except in cplex...

    def rename_linear_constraint(self, linct, new_name):
        pass  # nothing to do, except in cplex...

    def change_var_types(self, dvars, new_types):
        pass  # nothing to do, except in cplex...

    def create_linear_constraint(self, binaryct):
        return -1  # pragma: no cover

    def create_block_linear_constraints(self, ct_seq):
        return [-1] * len(ct_seq)  # pragma: no cover

    def create_batch_logical_constraints(self, logcts, is_equivalence):
        return [-1] * len(logcts)  # pragma: no cover

    def remove_constraint(self, ct):
        pass  # pragma: no cover

    def remove_constraints(self, cts):
        pass  # pragma: no cover

    def set_objective_sense(self, sense):
        pass  # pragma: no cover

    def set_objective_expr(self, new_objexpr, old_objexpr):
        pass  # pragma: no cover

    def set_multi_objective_exprs(self, new_multiobjexprs, old_multiobjexprs,
                                  priorities=None, weights=None, abstols=None, reltols=None, objnames=None):
        pass  # pragma: no cover

    def end(self):
        pass  # pragma: no cover

    def register_callback(self, cb):
        pass  # pragma: no cover

    def unregister_callback(self, cb):
        pass  # pragma: no cover

    def connect_progress_listeners(self, listeners, model):
        if listeners:
            model.warning("Progress listeners require CPLEX, not supported on engine {0}.".format(self.name))

    def disconnect_progress_listeners(self, listeners):
        pass  # pragma: no cover

    def solve(self, mdl, parameters, **kwargs):
        return None  # pragma: no cover

    def get_solve_status(self):
        return JobSolveStatus.UNKNOWN  # pragma: no cover

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):
        raise None  # pragma: no cover

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):
        raise None  # pragma: no cover

    def get_cplex(self):
        raise DOcplexException("No CPLEX is available.")  # pragma: no cover

    def clean_before_solve(self):
        pass  # pragma: no cover

    def update_objective(self, expr, event, *args):
        # nothing to do except for cplex
        pass  # pragma: no cover

    def update_constraint(self, ct, event, *args):
        pass  # pragma: no cover

    def supports_logical_constraints(self):
        return True, None

    def supports_multi_objective(self):
        return True, None  # pragma: no cover

    def check_var_indices(self, dvars):
        pass  # pragma: no cover

    def check_constraint_indices(self, cts, ctscope):
        pass  # pragma: no cover

    def create_sos(self, sos):
        pass  # pragma: no cover

    def clear_all_sos(self):
        pass  # pragma: no cover

    def get_basis(self, mdl):
        return None, None  # pragma: no cover

    def set_lp_start(self, var_stats, ct_stats):  # pragma: no cover
        raise DOcplexException('set_lp_start() requires CPLEX, not available for {0}'.format(self.name))

    def add_lazy_constraints(self, lazies):
        pass

    def clear_lazy_constraints(self):
        pass

    def add_user_cuts(self, lazies):
        pass

    def clear_user_cuts(self):
        pass

    def update_extra_constraint(self, lct, qualifier, *args):
        pass


# noinspection PyAbstractClass,PyUnusedLocal,PyMethodMayBeStatic
class IndexerEngine(DummyEngine):
    """
    An abstract engine facade which generates unique indices for variables, constraints
    """

    def __init__(self, initial_index=0):
        DummyEngine.__init__(self)
        self._initial_index = initial_index  # CPLEX indices start at 0, not 1
        self.__var_counter = self._initial_index
        self._ct_counter = self._initial_index

    def _increment_vars(self, size):
        self.__var_counter += size
        return self.__var_counter

    def _increment_cts(self, size):
        self._ct_counter += size
        return self._ct_counter

    def create_one_variable(self, vartype, lb, ub, name):
        old_count = self.__var_counter
        self._increment_vars(1)
        return old_count

    def create_variables(self, nb_vars, vartype, lb, ub, name):
        old_count = self.__var_counter
        new_count = self._increment_vars(nb_vars)
        return list(range(old_count, new_count))

    def create_multitype_variables(self, keys, vartypes, lbs, ubs, names):
        old_count = self.__var_counter
        new_count = self._increment_vars(len(keys))
        return list(range(old_count, new_count))

    def _create_one_ct(self):
        old_ct_count = self._ct_counter
        self._increment_cts(1)
        return old_ct_count

    def create_linear_constraint(self, binaryct):
        return self._create_one_ct()

    def create_batch_cts(self, ct_seq):
        old_ct_count = self._ct_counter
        size = sum(1 for _ in ct_seq)  # iterator is consumed
        self._increment_cts(size)
        return range(old_ct_count, self._ct_counter)

    def create_block_linear_constraints(self, ct_seq):
        return self.create_batch_cts(ct_seq)

    def create_range_constraint(self, rangect):
        return self._create_one_ct()

    def create_logical_constraint(self, logct, is_equivalence):
        return self._create_one_ct()

    def create_batch_logical_constraints(self, logcts, is_equivalence):
        return self.create_batch_cts(logcts)

    def create_quadratic_constraint(self, ind):
        return self._create_one_ct()

    def create_pwl_constraint(self, pwl_ct):
        return self._create_one_ct()

    def get_all_reduced_costs(self, mdl):
        return {}

    def get_all_dual_values(self, mdl):
        return {}

    def get_all_slack_values(self, mdl):
        return {CplexScope.LINEAR_CT_SCOPE: {},
                CplexScope.QUAD_CT_SCOPE: {},
                CplexScope.IND_CT_SCOPE: {}}

    def set_objective_sense(self, sense):
        pass

    def set_objective_expr(self, new_objexpr, old_objexpr):
        pass

    def set_multi_objective_exprs(self, new_multiobjexprs, old_multiobjexprs,
                                  priorities=None, weights=None, abstols=None, reltols=None, objnames=None):
        pass

    def set_parameter(self, parameter, value):
        """ Changes the parameter value in the engine.

        For this limited type of engine, nothing to do.

        """
        pass

    def get_parameter(self, parameter):
        """ Gets the current value of a parameter.

        Params:
         parameter: the parameter for which we query the value.

        """
        return parameter.get()


class NoSolveEngine(IndexerEngine):
    def get_solve_details(self):
        SolveDetails.make_fake_details(time=0, feasible=False)

    # INTERNAL: a dummy engine that cannot solve.

    # noinspection PyUnusedLocal
    def __init__(self, mdl, **kwargs):
        IndexerEngine.__init__(self)

    @property
    def name(self):
        return "nosolve"

    @staticmethod
    def _no_cplex_error(mdl, method_name):  # pragma: no cover
        mdl.fatal("No CPLEX DLL and no DOcplexcloud credentials: {0} is not available".format(method_name))

    def solve(self, mdl, parameters, **kwargs):  # pragma: no cover
        """
        This solver cannot solve. never ever.
        """
        self._no_cplex_error(mdl, method_name="solve")
        return None

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):  # pragma: no cover
        self._no_cplex_error(mdl, method_name="solve_relaxed")
        return None

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):  # pragma: no cover
        self._no_cplex_error(mdl, method_name="refine_conflict")
        return None

    @staticmethod
    def make_from_model(mdl):
        # used in pickle
        eng = NoSolveEngine(mdl)
        eng._increment_vars(mdl.number_of_variables)
        eng._increment_cts(mdl.number_of_constraints)
        return eng


# noinspection PyUnusedLocal
class ZeroSolveEngine(IndexerEngine):
    # INTERNAL: a dummy engine that says it can solve
    # but returns an all-zero solution.
    def __init__(self, mdl, **kwargs):
        IndexerEngine.__init__(self)  # pragma: no cover
        self._last_solved_parameters = None

    def show_parameters(self, params):
        if params is None:
            print("DEBUG> parameters: None")
        else:
            if params.has_nondefaults():
                print("DEBUG> parameters:")
                params.print_information(indent_level=8)  #
            else:
                print("DEBUG> parameters: defaults")

    @property
    def last_solved_parameters(self):
        return self._last_solved_parameters

    @property
    def name(self):
        return "zero_solve"

    @staticmethod
    def get_var_zero_solution(dvar):
        return max(0, dvar.lb)

    def solve(self, mdl, parameters, **kwargs):
        # remember last solved params
        self._last_solved_parameters = parameters.clone() if parameters is not None else None
        self.show_parameters(parameters)
        return self.make_zero_solution(mdl)

    def make_zero_solution(self, mdl):
        # return a feasible value: max of zero and the lower bound
        zlb_map = {v: self.get_var_zero_solution(v) for v in mdl.iter_variables() if v.lb != 0}
        obj = mdl.objective_expr.constant
        return SolveSolution(mdl, obj=obj, var_value_map=zlb_map, solved_by=self.name)  # pragma: no cover

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):
        params = parameters or mdl.parameters
        self._last_solved_parameters = params
        self.show_parameters(params)
        return self.make_zero_solution(mdl)

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):
        return None  # pragma: no cover

    def get_solve_details(self):
        return SolveDetails.make_fake_details(time=0, feasible=True)
